//
//  NSObject+Signals.m
//  Signals
//
//  Created by Fredrik Olsson on 2006-09-26.
//  Copyright 2006 __MyCompanyName__. All rights reserved.
//

#import "NSObject+Signals.h"


//	Categories can not add instance variables, so we use a hash map from
//	Core Foundation, with our instance pointers as keys.
static NSMapTable *mySignals = NULL;


/*
 *	A private class for holding a single signal slot.
 *	A signal slot holds a target object, and a selector to perform.
 *	The target is retained.
 */
@interface _NSSignalSlot : NSObject {
@private
	id _target;
	SEL _selector;
}

+ (_NSSignalSlot *)signalSlotWithTarget:(id)target selector:(SEL)selector;
- (id)initWithTarget:(id)target selector:(SEL)selector;
- (void)dealloc;

//	Used for removing signals.
- (BOOL)matchingTarget:(id)target;
//	Used for removing signals, a NULL selector matches all.
- (BOOL)matchingSelector:(SEL)selector;

//	Make the target object perform the selector.
- (void)perform;

// Make the target object perform the selector with a single argument.
- (void)performWithObject:(id)object;

@end


/*
 *	A private category to NSObject with some helper methods.
 */
@interface NSObject (PrivateSignals)

//	Get a dictionary with my signals, create and add if missing.
- (NSMutableDictionary *)mySignals;

//	Get a set with _NSSignalSlots for my signal, crewate and add if missing.
- (NSMutableSet *)targetSetForSignal:(id)signal;

//	Override dealloc to remove from hash map on deallocationm or else memory leak.
- (void)dealloc;

@end


/*
 *	AND HEREAFTER BE IMPLEMENTATIONS!
 */
 

@implementation NSObject (Signals)

- (void)connectSignal:(id)signal withTarget:(id)target selector:(SEL)selector
{
	[[self targetSetForSignal:signal] addObject:[_NSSignalSlot signalSlotWithTarget:target selector:selector]];
}

- (void)connectSignal:(NSString *)signal withTarget:(id)target
{
	[self connectSignal:signal withTarget:target selector:NSSelectorFromString(signal)];
}

- (void)removeSignal:(id)signal withTarget:(id)target selector:(SEL)selector
{
	NSMutableSet *set = [self targetSetForSignal:signal];
	NSMutableSet *removeSet = [NSMutableSet setWithCapacity:[set count]];
	NSEnumerator *slotsEnum = [set objectEnumerator];
	_NSSignalSlot *slot = nil;
	while (slot = [slotsEnum nextObject]) {
		if ([slot matchingTarget:target] && [slot matchingSelector:selector]) {
			[removeSet addObject:slot];
		}
	}
	[set minusSet:removeSet];
}

- (void)removeSignal:(id)signal withTarget:(id)target
{
	[self removeSignal:signal withTarget:target selector:NULL];
}

- (void)removeSignal:(id)signal
{
	[[self mySignals] removeObjectForKey:signal];
}

- (void)removeAllSignals
{
	if (mySignals) {
		NSMapRemove(mySignals, self);
	}
}

- (void)emitSignal:(id)signal
{
	NSSet *set = [self targetSetForSignal:signal];
	[set makeObjectsPerformSelector:@selector(permorm)];
}

- (void)emitSignal:(id)signal withObject:(id)object
{
	NSSet *set = [self targetSetForSignal:signal];
	[set makeObjectsPerformSelector:@selector(performWithObject:) withObject:object];
}

@end


@implementation _NSSignalSlot

+ (_NSSignalSlot *)signalSlotWithTarget:(id)target selector:(SEL)selector
{
	return [[[_NSSignalSlot alloc] initWithTarget:target selector:selector] autorelease];
}

- (id)initWithTarget:(id)target selector:(SEL)selector
{
	if (self = [super init]) {
		_target = [target retain];
		_selector = selector;
	}
	return self;
}

- (void)dealloc
{
	[_target release];
	[super dealloc];
}

- (BOOL)matchingTarget:(id)target
{
	return [target isEqual:_target];
}

- (BOOL)matchingSelector:(SEL)selector
{
	return selector ? selector == _selector : YES;
}

- (void)perform
{
	[_target performSelector:_selector];
}

- (void)performWithObject:(id)object
{
	[_target performSelector:_selector withObject:object];
}

@end


@implementation NSObject (PrivateSignals)

- (NSMutableDictionary *)mySignals
{
	if (!mySignals) {
		mySignals = NSCreateMapTable(NSNonOwnedPointerMapKeyCallBacks, NSObjectMapValueCallBacks, 32);
	}
	NSMutableDictionary *dict = NSMapGet(mySignals, self);
	if (!dict) {
		dict = [NSMutableDictionary dictionaryWithCapacity:8];
		NSMapInsert(mySignals, self, dict);
	}
	return dict;
}

- (NSMutableSet *)targetSetForSignal:(id)signal
{
	NSMutableDictionary *dict = [self mySignals];
	NSMutableSet *set = [dict objectForKey:signal];
	if (!set) {
		set = [NSMutableSet set];
		[dict setObject:set forKey:signal];
	}
	return set;
}

- (void)dealloc
{
	[self removeAllSignals];
	NSDeallocateObject(self);
}

@end
